#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "tinypy.h"

/*
 * File: Dict Functions for dealing with dictionaries.
 */
int
tp_lua_hash(void const * va, int l)
{
	uint8_t const  *v = (uint8_t const *) va;
	int 		i        , step = (l >> 5) + 1;
	int 		h = l + (l >= 4 ? *v : 0);
	for (i = l; i >= step; i -= step) {
		h = h ^ ((h << 5) + (h >> 2) + v[i - 1]);
	}
	return h;
}

void
_tp_dict_free(tp_vm * tp, _tp_dict * self)
{
	tp_free(tp, self->items);
	tp_free(tp, self);
}

/*
 * void _tp_dict_reset(_tp_dict *self) {
 * memset(self->items,0,self->alloc*sizeof(tp_item)); self->len = 0;
 * self->used = 0; self->cur = 0; }
 */
int
tp_hash(tp_vm * tp, tp_obj v)
{
	tp_obj 		vv;
	int 		r;
	int 		n;

	switch (v.type) {
	case TP_NONE:
		return 0;
	case TP_NUMBER:
		return tp_lua_hash(&v.number.val, sizeof(tp_num));
	case TP_STRING:
		return tp_lua_hash(v.string.val, v.string.len);
	case TP_DICT:
		return tp_lua_hash(&v.dict.val, sizeof(void *));
	case TP_LIST:
		r = v.list.val->len;
		for (n = 0; n < v.list.val->len; n++) {
			vv = v.list.val->items[n];
			r += (vv.type != TP_LIST) ?
				tp_hash(tp, v.list.val->items[n]) :
				tp_lua_hash(&vv.list.val, sizeof(void *));
		}
		return r;
	case TP_FNC:
		return tp_lua_hash(&v.fnc.info, sizeof(void *));
	case TP_DATA:
		return tp_lua_hash(&v.data.val, sizeof(void *));
	}
	tp_raise(0, tp_string("(tp_hash) TypeError: value unhashable"));
}

void
_tp_dict_hash_set(tp_vm * tp, _tp_dict * self, int hash, tp_obj k, tp_obj v)
{
	tp_item 	item;
	int 		n;
	int 		i;
	int		idx = hash & self->mask;

	for (i = idx; i < idx + self->alloc; i++) {
		n = i & self->mask;
		if (self->items[n].used > 0) {
			continue;
		}
		if (self->items[n].used == 0) {
			self->used += 1;
		}
		item.used = 1;
		item.hash = hash;
		item.key = k;
		item.val = v;
		self->items[n] = item;
		self->len += 1;
		return;
	}
	tp_raise(, tp_string("(_tp_dict_hash_set) RuntimeError: ?"));
}

void
_tp_dict_tp_realloc(tp_vm * tp, _tp_dict * self, int len)
{
	tp_item        *items = self->items;
	int 		i;
	int		alloc = self->alloc;

	len = _tp_max(8, len);
	self->items = (tp_item *) tp_malloc(tp, len * sizeof(tp_item));
	self->alloc = len;
	self->mask = len - 1;
	self->len = 0;
	self->used = 0;
	for (i = 0; i < alloc; i++) {
		if (items[i].used != 1) {
			continue;
		}
		_tp_dict_hash_set(tp, self, items[i].hash, items[i].key, items[i].val);
	}
	tp_free(tp, items);
}

int
_tp_dict_hash_find(tp_vm * tp, _tp_dict * self, int hash, tp_obj k)
{
	int 		i;
	int 		idx = hash & self->mask;
	int 		n;

	for (i = idx; i < idx + self->alloc; i++) {
		n = i & self->mask;
		if (self->items[n].used == 0) {
			break;
		}
		if (self->items[n].used < 0) {
			continue;
		}
		if (self->items[n].hash != hash) {
			continue;
		}
		if (tp_cmp(tp, self->items[n].key, k) != 0) {
			continue;
		}
		return n;
	}
	return -1;
}

int
_tp_dict_find(tp_vm * tp, _tp_dict * self, tp_obj k)
{
	return _tp_dict_hash_find(tp, self, tp_hash(tp, k), k);
}

void
_tp_dict_setx(tp_vm * tp, _tp_dict * self, tp_obj k, tp_obj v)
{
	int 		hash = tp_hash(tp, k);
	int 		n = _tp_dict_hash_find(tp, self, hash, k);

	if (n == -1) {
		if (self->len >= (self->alloc / 2)) {
			_tp_dict_tp_realloc(tp, self, self->alloc * 2);
		} else if (self->used >= (self->alloc * 3 / 4)) {
			_tp_dict_tp_realloc(tp, self, self->alloc);
		}
		_tp_dict_hash_set(tp, self, hash, k, v);
	} else {
		self->items[n].val = v;
	}
}

void
_tp_dict_set(tp_vm * tp, _tp_dict * self, tp_obj k, tp_obj v)
{
	_tp_dict_setx(tp, self, k, v);
	tp_grey(tp, k);
	tp_grey(tp, v);
}

tp_obj
_tp_dict_get(tp_vm * tp, _tp_dict * self, tp_obj k, const char *error)
{
	int 		n = _tp_dict_find(tp, self, k);

	if (n < 0) {
		tp_raise(tp_None, tp_add(tp, tp_string("(_tp_dict_get) KeyError: "), tp_str(tp, k)));
	}
	return self->items[n].val;
}

void
_tp_dict_del(tp_vm * tp, _tp_dict * self, tp_obj k, const char *error)
{
	int 		n = _tp_dict_find(tp, self, k);

	if (n < 0) {
		tp_raise(, tp_add(tp, tp_string("(_tp_dict_del) KeyError: "), tp_str(tp, k)));
	}
	self->items[n].used = -1;
	self->len -= 1;
}

_tp_dict       *
_tp_dict_new(tp_vm * tp)
{
	_tp_dict       *self = (_tp_dict *) tp_malloc(tp, sizeof(_tp_dict));

	return self;
}

tp_obj
_tp_dict_copy(tp_vm * tp, tp_obj rr)
{
	tp_obj 		obj = {TP_DICT};
	_tp_dict       *o = rr.dict.val;
	_tp_dict       *r = _tp_dict_new(tp);

	*r = *o;
	r->gci = 0;
	r->items = (tp_item *) tp_malloc(tp, sizeof(tp_item) * o->alloc);
	memcpy(r->items, o->items, sizeof(tp_item) * o->alloc);
	obj.dict.val = r;
	obj.dict.dtype = 1;
	return tp_track(tp, obj);
}

int
_tp_dict_next(tp_vm * tp, _tp_dict * self)
{
	if (!self->len) {
		tp_raise(0, tp_string("(_tp_dict_next) RuntimeError"));
	}
	for (;;) {
		self->cur = ((self->cur + 1) & self->mask);
		if (self->items[self->cur].used > 0) {
			return self->cur;
		}
	}
}

tp_obj
tp_merge(tp_vm * tp)
{
	tp_obj 		self = TP_OBJ();
	tp_obj 		v = TP_OBJ();
	int 		i;

	for (i = 0; i < v.dict.val->len; i++) {
		int 		n = _tp_dict_next(tp, v.dict.val);
		_tp_dict_set(tp, self.dict.val,
			v.dict.val->items[n].key, v.dict.val->items[n].val);
	}
	return tp_None;
}

/*
 * Function: tp_dict
 *
 * Creates a new dictionary object.
 *
 *Note* If you use <tp_setmeta> on the dictionary, you have to use <tp_getraw>
 * to access the "raw" dictionary again.
 *
 * Returns: The newly created dictionary.
 */
tp_obj
tp_dict(tp_vm * tp)
{
	tp_obj 		r = {TP_DICT};

	r.dict.val = _tp_dict_new(tp);
	r.dict.dtype = 1;
	return tp ? tp_track(tp, r) : r;
}

tp_obj
tp_dict_n(tp_vm * tp, int n, tp_obj * argv)
{
	tp_obj 		r = tp_dict(tp);
	int 		i;

	for (i = 0; i < n; i++) {
		tp_set(tp, r, argv[i * 2], argv[i * 2 + 1]);
	}
	return r;
}
