/*
================================================================================

tinypy contains tinypy code licensed in a MIT format license.  It also
contains some goodies grabbed from Python, so that license is included
as well.

================================================================================

The tinypy License

Copyright (c) 2008 Phil Hassey

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

================================================================================

PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
--------------------------------------------

1. This LICENSE AGREEMENT is between the Python Software Foundation
("PSF"), and the Individual or Organization ("Licensee") accessing and
otherwise using this software ("Python") in source or binary form and
its associated documentation.

2. Subject to the terms and conditions of this License Agreement, PSF
hereby grants Licensee a nonexclusive, royalty-free, world-wide
license to reproduce, analyze, test, perform and/or display publicly,
prepare derivative works, distribute, and otherwise use Python
alone or in any derivative version, provided, however, that PSF's
License Agreement and PSF's notice of copyright, i.e., "Copyright (c)
2001, 2002, 2003, 2004, 2005, 2006, 2007 Python Software Foundation;
All Rights Reserved" are retained in Python alone or in any derivative
version prepared by Licensee.

3. In the event Licensee prepares a derivative work that is based on
or incorporates Python or any part thereof, and wants to make
the derivative work available to others as provided herein, then
Licensee hereby agrees to include in any such work a brief summary of
the changes made to Python.

4. PSF is making Python available to Licensee on an "AS IS"
basis.  PSF MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR
IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND
DISCLAIMS ANY REPRESENTATION OR WARRANTY OF MERCHANTABILITY OR FITNESS
FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT
INFRINGE ANY THIRD PARTY RIGHTS.

5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON
FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL DAMAGES OR LOSS AS
A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON,
OR ANY DERIVATIVE THEREOF, EVEN IF ADVISED OF THE POSSIBILITY THEREOF.

6. This License Agreement will automatically terminate upon a material
breach of its terms and conditions.

7. Nothing in this License Agreement shall be deemed to create any
relationship of agency, partnership, or joint venture between PSF and
Licensee.  This License Agreement does not grant permission to use PSF
trademarks or trade name in a trademark sense to endorse or promote
products or services of Licensee, or any third party.

8. By copying, installing or otherwise using Python, Licensee
agrees to be bound by the terms and conditions of this License
Agreement.

================================================================================
*/

#ifndef TINYPY_H
#define TINYPY_H

#include <inttypes.h>
#include <setjmp.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#define TP_INLINE		__inline

enum {
	TP_NONE			= 0,
	TP_NUMBER		= 1,
	TP_STRING		= 2,
	TP_DICT			= 3,
	TP_LIST			= 4,
	TP_FNC			= 5,
	TP_DATA			= 6,
};

#ifndef TP_NUMBER_TYPE
#define TP_NUMBER_TYPE		 double
#endif

typedef TP_NUMBER_TYPE		 tp_num;

typedef struct tp_number_ {
	int			 type;
	tp_num			 val;
} tp_number_;

typedef struct tp_string_ {
	int			 type;
	struct _tp_string	*info;
	char const		*val;
	int			 len;
} tp_string_;

typedef struct tp_list_ {
	int			 type;
	struct _tp_list		*val;
} tp_list_;

typedef struct tp_dict_ {
	int			 type;
	struct _tp_dict		*val;
	int			 dtype;
} tp_dict_;

#define TP_TYPE_FUNC		0
#define TP_TYPE_DEF		1
#define TP_TYPE_METHOD		2

typedef struct tp_fnc_ {
	int			 type;
	struct _tp_fnc		*info;
	int			 ftype;
	void			*cfnc;
} tp_fnc_;

typedef struct tp_data_ {
	int			 type;
	struct _tp_data		*info;
	void			*val;
	int			 magic;
} tp_data_;

/* Type: tp_obj
 * Tinypy's object representation.
 *
 * Every object in tinypy is of this type in the C API.
 *
 * Fields:
 * type - This determines what kind of objects it is. It is either TP_NONE, in
 *        which case this is the none type and no other fields can be accessed.
 *        Or it has one of the values listed below, and the corresponding
 *        fields can be accessed.
 * number - TP_NUMBER
 * number.val - A double value with the numeric value.
 * string - TP_STRING
 * string.val - A pointer to the string data.
 * string.len - Length in bytes of the string data.
 * dict - TP_DICT
 * list - TP_LIST
 * fnc - TP_FNC
 * data - TP_DATA
 * data.val - The user-provided data pointer.
 * data.magic - The user-provided magic number for identifying the data type.
 */
typedef union tp_obj {
	int		 type;
	tp_number_	 number;
	struct {
		int	 type;
		int	*data;
	} gci;
	tp_string_	 string;
	tp_dict_	 dict;
	tp_list_	 list;
	tp_fnc_		 fnc;
	tp_data_	 data;
} tp_obj;

typedef struct _tp_string {
	int		 gci;
	int		 len;
	char		 s[1];
} _tp_string;

typedef struct _tp_list {
	int		 gci;
	tp_obj		*items;
	int		 len;
	int		 alloc;
} _tp_list;

typedef struct tp_item {
	int		 used;
	int		 hash;
	tp_obj		 key;
	tp_obj		 val;
} tp_item;

typedef struct _tp_dict {
	int		 gci;
	tp_item		*items;
	int		 len;
	int		 alloc;
	int		 cur;
	int		 mask;
	int		 used;
	tp_obj		 meta;
} _tp_dict;

typedef struct _tp_fnc {
	int		 gci;
	tp_obj		 self;
	tp_obj		 globals;
	tp_obj		 code;
} _tp_fnc;

typedef union tp_code {
	uint8_t			 i;
	struct {
		uint8_t	 i;
		uint8_t	 a;
		uint8_t	 b;
		uint8_t	 c;
	}			 regs;
	struct {
		char	 val[4];
	}			 string;
	struct {
		float	 val;
	}			 number;
} tp_code;

typedef struct tp_frame_ {
/*    tp_code *codes; */
	tp_obj			 code;
	tp_code			*in;		/* current vm instruction */
	tp_code			*jmp;
	tp_obj			*regs;
	tp_obj			*ret_dest;
	tp_obj			 fname;
	tp_obj			 name;
	tp_obj			 line;
	tp_obj			 globals;
	int			 lineno;
	int			 cregs;
} tp_frame_;

#define TP_GCMAX 4096
#define TP_FRAMES 256
#define TP_REGS_EXTRA 2
/* #define TP_REGS_PER_FRAME 256*/
#define TP_REGS 16384

/* Type: tp_vm
 * Representation of a tinypy virtual machine instance.
 *
 * A new tp_vm struct is created with <tp_init>, and will be passed to most
 * tinypy functions as first parameter. It contains all the data associated
 * with an instance of a tinypy virtual machine - so it is easy to have
 * multiple instances running at the same time. When you want to free up all
 * memory used by an instance, call <tp_deinit>.
 *
 * Fields:
 * These fields are currently documented:
 *
 * builtins - A dictionary containing all builtin objects.
 * modules - A dictionary with all loaded modules.
 * params - A list of parameters for the current function call.
 * frames - A list of all call frames.
 * cur - The index of the currently executing call frame.
 * frames[n].globals - A dictionary of global sybmols in callframe n.
 */
typedef struct tp_vm {
	tp_obj		 builtins;
	tp_obj		 modules;
	tp_frame_	 frames[TP_FRAMES];
	tp_obj		 _params;
	tp_obj		 params;
	tp_obj		 _regs;
	tp_obj		*regs;
	tp_obj		 root;
	jmp_buf		 buf;
#ifdef CPYTHON_MOD
	jmp_buf		 nextexpr;
#endif
	int		 jmp;
	tp_obj		 ex;			/* exit code */
	char		 chars[256][2];
	int		 frame;			/* current frame */
	/* gc */
	_tp_list	*white;
	_tp_list	*grey;
	_tp_list	*black;
	int		 steps;
	/* sandbox */
	clock_t		 clocks;
	TP_NUMBER_TYPE	 time_elapsed;
	TP_NUMBER_TYPE	 time_limit;
	uint64_t	 mem_limit;
	uint64_t	 mem_used;
	int		 mem_exceeded;
} tp_vm;

typedef struct _tp_data {
	int gci;
	void (*free)(tp_vm *tp,tp_obj);
} _tp_data;

#define tp_True tp_number(1)
#define tp_False tp_number(0)

extern tp_obj tp_None;

#ifndef __BEGIN_DECLS
#  if defined(__cplusplus)
#  define __BEGIN_DECLS           extern "C" {
#  define __END_DECLS             }
#  else
#  define __BEGIN_DECLS
#  define __END_DECLS
#  endif
#endif

__BEGIN_DECLS

#ifdef TP_SANDBOX
void *tp_malloc(tp_vm */*tp*/, uint64_t);
void *tp_realloc(tp_vm */*tp*/, void *, uint64_t);
void tp_free(tp_vm */*tp*/, void *);
#else
#define tp_malloc(Y,x) calloc((x),1)
#define tp_realloc(Y,x,y) realloc(x,y)
#define tp_free(Y,x) free(x)
#endif

void tp_sandbox(tp_vm */*tp*/, uint64_t, uint64_t);
void tp_time_update(tp_vm */*tp*/);
void tp_mem_update(tp_vm */*tp*/);

void tp_run(tp_vm */*tp*/,int /*cur*/);
void tp_set(tp_vm */*tp*/,tp_obj,tp_obj,tp_obj);
tp_obj tp_get(tp_vm */*tp*/,tp_obj,tp_obj);
tp_obj tp_has(tp_vm */*tp*/,tp_obj /*self*/, tp_obj /*k*/);
tp_obj tp_len(tp_vm */*tp*/,tp_obj);
void tp_del(tp_vm */*tp*/,tp_obj,tp_obj);
tp_obj tp_str(tp_vm */*tp*/,tp_obj);
int tp_bool(tp_vm */*tp*/,tp_obj);
int tp_cmp(tp_vm */*tp*/,tp_obj,tp_obj);
void _tp_raise(tp_vm */*tp*/,tp_obj);
tp_obj tp_printf(tp_vm */*tp*/,char const */*fmt*/,...);
tp_obj tp_track(tp_vm */*tp*/,tp_obj);
void tp_grey(tp_vm */*tp*/,tp_obj);
tp_obj tp_call(tp_vm */*tp*/, tp_obj /*fnc*/, tp_obj /*params*/);
tp_obj tp_add(tp_vm */*tp*/,tp_obj /*a*/, tp_obj /*b*/);

/* __func__ __VA_ARGS__ __FILE__ __LINE__ */

/* Function: tp_raise
 * Macro to raise an exception.
 *
 * This macro will return from the current function returning "r". The
 * remaining parameters are used to format the exception message.
 */
/*
#define tp_raise(r,fmt,...) { \
    _tp_raise(tp,tp_printf(tp,fmt,__VA_ARGS__)); \
    return r; \
}
*/
#define tp_raise(r,v) { \
    _tp_raise(tp,v); \
    return r; \
}

/* Function: tp_string
 * Creates a new string object from a C string.
 *
 * Given a pointer to a C string, creates a tinypy object representing the
 * same string.
 *
 * *Note* Only a reference to the string will be kept by tinypy, so make sure
 * it does not go out of scope, and don't de-allocate it. Also be aware that
 * tinypy will not delete the string for you. In many cases, it is best to
 * use <tp_string_t> or <tp_string_slice> to create a string where tinypy
 * manages storage for you.
 */
static TP_INLINE tp_obj tp_string(char const *v)
{
    tp_obj val;
    tp_string_ s = {TP_STRING, 0, v, 0};
    s.len = strlen(v);
    val.string = s;
    return val;
}

#define TP_CSTR_LEN 256

static TP_INLINE void tp_cstr(tp_vm *tp,tp_obj v, char *s, int l)
{
	if (v.type != TP_STRING) {
		tp_raise(,tp_string("(tp_cstr) TypeError: value not a string"));
	}
	if (v.string.len >= l) {
		tp_raise(,tp_string("(tp_cstr) TypeError: value too long"));
	}
	memset(s, 0, l);
	memcpy(s, v.string.val, v.string.len);
}

#define TP_OBJ() (tp_get(tp, tp->params, tp_None))

static TP_INLINE tp_obj tp_type(tp_vm *tp, int t, tp_obj v)
{
	if (v.type != t) {
		tp_raise(tp_None, tp_string("(tp_type) TypeError: unexpected type"));
	}
	return v;
}

#define TP_NO_LIMIT 0
#define TP_TYPE(t) tp_type(tp,t,TP_OBJ())
#define TP_NUM() (TP_TYPE(TP_NUMBER).number.val)
/* #define TP_STR() (TP_CSTR(TP_TYPE(TP_STRING))) */
#define TP_STR() (TP_TYPE(TP_STRING))
#define TP_DEFAULT(d) (tp->params.list.val->len?tp_get(tp,tp->params,tp_None):(d))

static TP_INLINE int _tp_min(int a, int b) { return (a<b?a:b); }
static TP_INLINE int _tp_max(int a, int b) { return (a>b?a:b); }
static TP_INLINE int _tp_sign(tp_num v) { return (v<0?-1:(v>0?1:0)); }

/* Function: tp_number
 * Creates a new numeric object.
 */
static TP_INLINE tp_obj tp_number(tp_num v)
{
	tp_obj	val = {TP_NUMBER};

	val.number.val = v;
	return val;
}

static TP_INLINE void tp_echo(tp_vm *tp,tp_obj e)
{
	e = tp_str(tp,e);
	fwrite(e.string.val,1,e.string.len,stdout);
}

/* Function: tp_string_n
 * Creates a new string object from a partial C string.
 *
 * Like <tp_string>, but you specify how many bytes of the given C string to
 * use for the string object. The *note* also applies for this function, as the
 * string reference and length are kept, but no actual substring is stored.
 */
static TP_INLINE tp_obj tp_string_n(char const *v,int n)
{
	tp_obj		val;
	tp_string_	s = {TP_STRING, 0, v, n};

	val.string = s;
	return val;
}

void _tp_list_realloc(tp_vm */*tp*/, _tp_list */*self*/,int /*len*/);
void _tp_list_set(tp_vm */*tp*/,_tp_list */*self*/,int /*k*/, tp_obj /*v*/, const char */*error*/);
void _tp_list_free(tp_vm */*tp*/, _tp_list */*self*/);
tp_obj _tp_list_get(tp_vm */*tp*/,_tp_list */*self*/,int /*k*/,const char */*error*/);
void _tp_list_insertx(tp_vm */*tp*/,_tp_list */*self*/, int /*n*/, tp_obj /*v*/);
void _tp_list_appendx(tp_vm */*tp*/,_tp_list */*self*/, tp_obj /*v*/);
void _tp_list_insert(tp_vm */*tp*/,_tp_list */*self*/, int /*n*/, tp_obj /*v*/);
void _tp_list_append(tp_vm */*tp*/,_tp_list */*self*/, tp_obj /*v*/);
tp_obj _tp_list_pop(tp_vm */*tp*/,_tp_list */*self*/, int /*n*/, const char */*error*/);
int _tp_list_find(tp_vm */*tp*/,_tp_list */*self*/, tp_obj /*v*/);
tp_obj tp_index(tp_vm */*tp*/);
_tp_list *_tp_list_new(tp_vm */*tp*/);
tp_obj _tp_list_copy(tp_vm */*tp*/, tp_obj /*rr*/);
tp_obj tp_append(tp_vm */*tp*/);
tp_obj tp_pop(tp_vm */*tp*/);
tp_obj tp_insert(tp_vm */*tp*/);
tp_obj tp_extend(tp_vm */*tp*/);
tp_obj tp_list_nt(tp_vm */*tp*/);
tp_obj tp_list(tp_vm */*tp*/);
tp_obj tp_list_n(tp_vm */*tp*/,int /*n*/,tp_obj */*argv*/);
int _tp_sort_cmp(tp_obj */*a*/,tp_obj */*b*/);
tp_obj tp_sort(tp_vm */*tp*/);
int tp_lua_hash(void const */*v*/,int /*l*/);
void _tp_dict_free(tp_vm */*tp*/, _tp_dict */*self*/);
int tp_hash(tp_vm */*tp*/,tp_obj /*v*/);
void _tp_dict_hash_set(tp_vm */*tp*/,_tp_dict */*self*/, int /*hash*/, tp_obj /*k*/, tp_obj /*v*/);
void _tp_dict_tp_realloc(tp_vm */*tp*/,_tp_dict */*self*/,int /*len*/);
int _tp_dict_hash_find(tp_vm */*tp*/,_tp_dict */*self*/, int /*hash*/, tp_obj /*k*/);
int _tp_dict_find(tp_vm */*tp*/,_tp_dict */*self*/,tp_obj /*k*/);
void _tp_dict_setx(tp_vm */*tp*/,_tp_dict */*self*/,tp_obj /*k*/, tp_obj /*v*/);
void _tp_dict_set(tp_vm */*tp*/,_tp_dict */*self*/,tp_obj /*k*/, tp_obj /*v*/);
tp_obj _tp_dict_get(tp_vm */*tp*/,_tp_dict */*self*/,tp_obj /*k*/, const char */*error*/);
void _tp_dict_del(tp_vm */*tp*/,_tp_dict */*self*/,tp_obj /*k*/, const char */*error*/);
_tp_dict *_tp_dict_new(tp_vm */*tp*/);
tp_obj _tp_dict_copy(tp_vm */*tp*/,tp_obj /*rr*/);
int _tp_dict_next(tp_vm */*tp*/,_tp_dict */*self*/);
tp_obj tp_merge(tp_vm */*tp*/);
tp_obj tp_dict(tp_vm */*tp*/);
tp_obj tp_dict_n(tp_vm */*tp*/,int /*n*/, tp_obj* /*argv*/);
tp_obj _tp_dcall(tp_vm */*tp*/,tp_obj fnc(tp_vm */*tp*/));
tp_obj _tp_tcall(tp_vm */*tp*/,tp_obj /*fnc*/);
tp_obj tp_fnc_new(tp_vm */*tp*/,int /*t*/, void */*v*/, tp_obj /*c*/,tp_obj /*s*/, tp_obj /*g*/);
tp_obj tp_def(tp_vm */*tp*/,tp_obj /*code*/, tp_obj /*g*/);
tp_obj tp_fnc(tp_vm */*tp*/,tp_obj v(tp_vm */*tp*/));
tp_obj tp_method(tp_vm */*tp*/,tp_obj /*self*/,tp_obj v(tp_vm */*tp*/));
tp_obj tp_data(tp_vm */*tp*/,int /*magic*/,void */*v*/);
tp_obj tp_params(tp_vm */*tp*/);
tp_obj tp_params_n(tp_vm */*tp*/,int /*n*/, tp_obj /*argv*/[]);
tp_obj tp_params_v(tp_vm */*tp*/,int /*n*/,...);
tp_obj tp_string_t(tp_vm */*tp*/, int /*n*/);
tp_obj tp_string_t2(tp_vm */*tp*/, void */*v*/, int /*n*/);
tp_obj tp_string_copy(tp_vm */*tp*/, const char */*s*/, int /*n*/);
tp_obj tp_string_sub(tp_vm */*tp*/, tp_obj /*s*/, int /*a*/, int /*b*/);
int _tp_str_index(tp_obj /*s*/, tp_obj /*k*/);
tp_obj tp_join(tp_vm */*tp*/);
tp_obj tp_split(tp_vm */*tp*/);
tp_obj tp_find(tp_vm */*tp*/);
tp_obj tp_str_index(tp_vm */*tp*/);
tp_obj tp_builtin_str(tp_vm */*tp*/);
tp_obj tp_chr(tp_vm */*tp*/);
tp_obj tp_ord(tp_vm */*tp*/);
tp_obj tp_strip(tp_vm */*tp*/);
tp_obj tp_replace(tp_vm */*tp*/);
tp_obj tp_print(tp_vm */*tp*/);
tp_obj tp_bind(tp_vm */*tp*/);
tp_obj tp_min(tp_vm */*tp*/);
tp_obj tp_max(tp_vm */*tp*/);
tp_obj tp_copy(tp_vm */*tp*/);
tp_obj tp_builtin_len(tp_vm */*tp*/);
tp_obj tp_assert(tp_vm */*tp*/);
tp_obj tp_range(tp_vm */*tp*/);
tp_obj tp_system(tp_vm */*tp*/);
tp_obj tp_istype(tp_vm */*tp*/);
tp_obj tp_float(tp_vm */*tp*/);
tp_obj tp_save(tp_vm */*tp*/);
tp_obj tp_load(tp_vm */*tp*/);
tp_obj tp_fpack(tp_vm */*tp*/);
tp_obj tp_abs(tp_vm */*tp*/);
tp_obj tp_int(tp_vm */*tp*/);
tp_num _roundf(tp_num v);
tp_obj tp_round(tp_vm */*tp*/);
tp_obj tp_exists(tp_vm */*tp*/);
tp_obj tp_mtime(tp_vm */*tp*/);
int _tp_lookup_(tp_vm */*tp*/,tp_obj /*self*/, tp_obj /*k*/, tp_obj */*meta*/, int /*depth*/);
int _tp_lookup(tp_vm */*tp*/,tp_obj /*self*/, tp_obj /*k*/, tp_obj */*meta*/);
tp_obj tp_setmeta(tp_vm */*tp*/);
tp_obj tp_getmeta(tp_vm */*tp*/);
tp_obj tp_object(tp_vm */*tp*/);
tp_obj tp_object_new(tp_vm */*tp*/);
tp_obj tp_object_call(tp_vm */*tp*/);
tp_obj tp_getraw(tp_vm */*tp*/);
tp_obj tp_class(tp_vm */*tp*/);
tp_obj tp_builtins_bool(tp_vm */*tp*/);
void tp_follow(tp_vm */*tp*/,tp_obj /*v*/);
void tp_reset(tp_vm */*tp*/);
void tp_gc_init(tp_vm */*tp*/);
void tp_gc_deinit(tp_vm */*tp*/);
void tp_delete(tp_vm */*tp*/,tp_obj /*v*/);
void tp_collect(tp_vm */*tp*/);
void _tp_gcinc(tp_vm */*tp*/);
void tp_full(tp_vm */*tp*/);
void tp_gcinc(tp_vm */*tp*/);
tp_obj tp_iter(tp_vm */*tp*/,tp_obj /*self*/, tp_obj /*k*/);
int tp_iget(tp_vm */*tp*/,tp_obj */*r*/, tp_obj /*self*/, tp_obj /*k*/);
tp_obj tp_mul(tp_vm */*tp*/,tp_obj /*a*/, tp_obj /*b*/);
tp_obj tp_bitwise_not(tp_vm */*tp*/, tp_obj /*a*/);
void tp_deinit(tp_vm */*tp*/);
void tp_frame(tp_vm */*tp*/,tp_obj /*globals*/,tp_obj /*code*/,tp_obj */*ret_dest*/);
void tp_print_stack(tp_vm */*tp*/);
void tp_handle(tp_vm */*tp*/);
void tp_return(tp_vm */*tp*/, tp_obj /*v*/);
int tp_step(tp_vm */*tp*/);
tp_obj tp_ez_call(tp_vm */*tp*/, const char */*mod*/, const char */*fnc*/, tp_obj /*params*/);
tp_obj tp_import(tp_vm */*tp*/, const char * /*fname*/, const char * /*name*/, void */*codes*/, int /*len*/);
tp_obj tp_builtin_exec(tp_vm */*tp*/);
tp_obj tp_builtin_import(tp_vm */*tp*/);
void tp_builtins(tp_vm */*tp*/);
void tp_args(tp_vm */*tp*/,int /*argc*/, char */*argv*/[], int /*start*/);
tp_obj tp_main(tp_vm */*tp*/,char */*fname*/, void */*code*/, int /*len*/);
tp_obj tp_compile(tp_vm */*tp*/, tp_obj /*text*/, tp_obj /*fname*/);
tp_obj tp_exec(tp_vm */*tp*/, tp_obj /*code*/, tp_obj /*globals*/);
tp_obj tp_eval(tp_vm */*tp*/, const char */*text*/, tp_obj /*globals*/);
tp_vm *tp_init(int /*argc*/, char */*argv*/[], int /*start*/);
int tp_compiler(tp_vm */*tp*/);
tp_obj tp_builtin_sandbox(tp_vm */*tp*/);
void tp_bounds(tp_vm */*tp*/, tp_code */*cur*/, int /*n*/);

tp_obj tp_bitwise_and(tp_vm */*tp*/, tp_obj /*_a*/, tp_obj /*_b*/);
tp_obj tp_bitwise_or(tp_vm */*tp*/, tp_obj /*_a*/, tp_obj /*_b*/);
tp_obj tp_bitwise_xor(tp_vm */*tp*/, tp_obj /*_a*/, tp_obj /*_b*/);
tp_obj tp_mod(tp_vm */*tp*/, tp_obj /*_a*/, tp_obj /*_b*/);
tp_obj tp_lsh(tp_vm */*tp*/, tp_obj /*_a*/, tp_obj /*_b*/);
tp_obj tp_rsh(tp_vm */*tp*/, tp_obj /*_a*/, tp_obj /*_b*/);
tp_obj tp_sub(tp_vm */*tp*/, tp_obj /*_a*/, tp_obj /*_b*/);
tp_obj tp_div(tp_vm */*tp*/, tp_obj /*_a*/, tp_obj /*_b*/);
tp_obj tp_pow(tp_vm */*tp*/, tp_obj /*_a*/, tp_obj /*_b*/);
tp_obj tp_bitwise_not(tp_vm */*tp*/, tp_obj /*_a*/);

__END_DECLS

#endif
