#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "tinypy.h"

/*
 * File: VM Functionality pertaining to the virtual machine.
 */

static tp_vm *
internal_init(void)
{
	int 		i;
	tp_vm          *tp = (tp_vm *) calloc(sizeof(tp_vm), 1);
	tp_obj 		sys;

	tp->time_limit = TP_NO_LIMIT;
	tp->clocks = clock();
	tp->time_elapsed = 0.0;
	tp->mem_limit = TP_NO_LIMIT;
	tp->mem_exceeded = 0;
	tp->mem_used = sizeof(tp_vm);
	tp->frame = 0;
	tp->jmp = 0;
	tp->ex = tp_None;
	tp->root = tp_list_nt(tp);
	for (i = 0; i < 256; i++) {
		tp->chars[i][0] = i;
	}
	tp_gc_init(tp);
	tp->_regs = tp_list(tp);
	for (i = 0; i < TP_REGS; i++) {
		tp_set(tp, tp->_regs, tp_None, tp_None);
	}
	tp->builtins = tp_dict(tp);
	tp->modules = tp_dict(tp);
	tp->_params = tp_list(tp);
	for (i = 0; i < TP_FRAMES; i++) {
		tp_set(tp, tp->_params, tp_None, tp_list(tp));
	}
	tp_set(tp, tp->root, tp_None, tp->builtins);
	tp_set(tp, tp->root, tp_None, tp->modules);
	tp_set(tp, tp->root, tp_None, tp->_regs);
	tp_set(tp, tp->root, tp_None, tp->_params);
	tp_set(tp, tp->builtins, tp_string("MODULES"), tp->modules);
	tp_set(tp, tp->modules, tp_string("BUILTINS"), tp->builtins);
	tp_set(tp, tp->builtins, tp_string("BUILTINS"), tp->builtins);
	sys = tp_dict(tp);
	tp_set(tp, sys, tp_string("version"), tp_string("tinypy 1.2+SVN"));
	tp_set(tp, tp->modules, tp_string("sys"), sys);
	tp->regs = tp->_regs.list.val->items;
	tp_full(tp);
	return tp;
}


/*
 * Function: tp_deinit Destroys a VM instance.
 * 
 * When you no longer need an instance of tinypy, you can use this to free all
 * memory used by it. Even when you are using only a single tinypy instance,
 * it may be good practice to call this function on shutdown.
 */
void 
tp_deinit(tp_vm *tp)
{
	while (tp->root.list.val->len) {
		_tp_list_pop(tp, tp->root.list.val, 0, "tp_deinit");
	}
	tp_full(tp);
	tp_full(tp);
	tp_delete(tp, tp->root);
	tp_gc_deinit(tp);
	tp->mem_used -= sizeof(tp_vm);
	free(tp);
}

/* tp_frame_ */
void 
tp_frame(tp_vm *tp, tp_obj globals, tp_obj code, tp_obj *ret_dest)
{
	tp_frame_ 	f;

	f.globals = globals;
	f.code = code;
	f.in = __UNCONST(f.code.string.val);
	f.jmp = 0;
	/* fprintf(stderr,"tp->frame: %d\n",tp->frame); */
	f.regs = (tp->frame <= 0 ? tp->regs : tp->frames[tp->frame].regs + tp->frames[tp->frame].cregs);

	f.regs[0] = f.globals;
	f.regs[1] = f.code;
	f.regs += TP_REGS_EXTRA;

	f.ret_dest = ret_dest;
	f.lineno = 0;
	f.line = tp_string("");
	f.name = tp_string("?");
	f.fname = tp_string("?");
	f.cregs = 0;
	/* return f; */
	if (f.regs + (256 + TP_REGS_EXTRA) >= tp->regs + TP_REGS || tp->frame >= TP_FRAMES - 1) {
		tp_raise(, tp_string("(tp_frame) RuntimeError: stack overflow"));
	}
	tp->frame += 1;
	tp->frames[tp->frame] = f;
}

void 
_tp_raise(tp_vm *tp, tp_obj e)
{
	/* char *x = 0; x[0]=0; */
	if (!tp || !tp->jmp) {
#ifndef CPYTHON_MOD
		printf("\nException:\n");
		tp_echo(tp, e);
		printf("\n");
		exit(-1);
#else
		tp->ex = e;
		longjmp(tp->nextexpr, 1);
#endif
	}
	if (e.type != TP_NONE) {
		tp->ex = e;
	}
	tp_grey(tp, e);
	longjmp(tp->buf, 1);
}

void 
tp_print_stack(tp_vm *tp)
{
	int 		i;

	printf("\n");
	for (i = 0; i <= tp->frame; i++) {
		if (!tp->frames[i].lineno) {
			continue;
		}
		printf("File \"");
		tp_echo(tp, tp->frames[i].fname);
		printf("\", ");
		printf("line %d, in ", tp->frames[i].lineno);
		tp_echo(tp, tp->frames[i].name);
		printf("\n ");
		tp_echo(tp, tp->frames[i].line);
		printf("\n");
	}
	printf("\nException:\n");
	tp_echo(tp, tp->ex);
	printf("\n");
}

void 
tp_handle(tp_vm *tp)
{
	int 		i;

	for (i = tp->frame; i >= 0; i--) {
		if (tp->frames[i].jmp) {
			break;
		}
	}
	if (i >= 0) {
		tp->frame = i;
		tp->frames[i].in = tp->frames[i].jmp;
		tp->frames[i].jmp = 0;
		return;
	}
#ifndef CPYTHON_MOD
	tp_print_stack(tp);
	exit(-1);
#else
	longjmp(tp->nextexpr, 1);
#endif
}

/*
 * Function: tp_call Calls a tinypy function.
 * 
 * Use this to call a tinypy function.
 * 
 * Parameters:
 *	tp - The VM instance.
 *	self - The object to call.
 *	params - Parameters to pass.
 * 
 * Example:
 *  > tp_call(tp,
 *  >     tp_get(tp, tp->builtins, tp_string("foo")),
 *  >     tp_params_v(tp, tp_string("hello")))
 *
 * This will look for a global function named "foo", then call it with
 * a single positional parameter containing the string "hello".
 */
tp_obj 
tp_call(tp_vm *tp, tp_obj self, tp_obj params)
{
	/*
	 * I'm not sure we should have to do this, but just for giggles we
	 * will.
	 */
	tp->params = params;

	if (self.type == TP_DICT) {
		if (self.dict.dtype == 1) {
			tp_obj 		meta;
			if (_tp_lookup(tp, self, tp_string("__new__"), &meta)) {
				_tp_list_insert(tp, params.list.val, 0, self);
				return tp_call(tp, meta, params);
			}
		} else if (self.dict.dtype == 2) {
			if (self.dict.dtype == 2) {
				tp_obj	meta;
				if (_tp_lookup(tp, self, tp_string("__call__"), &meta)) {
					return tp_call(tp, meta, params);
				}
			}
		}
	}
	if (self.type == TP_FNC && !(self.fnc.ftype & TP_TYPE_DEF)) {
		tp_obj 		r = _tp_tcall(tp, self);
		tp_grey(tp, r);
		return r;
	}
	if (self.type == TP_FNC) {
		tp_obj 		dest = tp_None;
		tp_frame(tp, self.fnc.info->globals, self.fnc.info->code, &dest);
		if ((self.fnc.ftype & TP_TYPE_METHOD)) {
			tp->frames[tp->frame].regs[0] = params;
			_tp_list_insert(tp, params.list.val, 0, self.fnc.info->self);
		} else {
			tp->frames[tp->frame].regs[0] = params;
		}
		tp_run(tp, tp->frame);
		return dest;
	}
	tp_params_v(tp, 1, self);
	tp_print(tp);
	tp_raise(tp_None, tp_string("(tp_call) TypeError: object is not callable"));
}


void 
tp_return(tp_vm * tp, tp_obj v)
{
	tp_obj         *dest = tp->frames[tp->frame].ret_dest;

	if (dest) {
		*dest = v;
		tp_grey(tp, v);
	}
	/*
	 * memset(tp->frames[tp->frame].regs,0,TP_REGS_PER_FRAME*sizeof(tp_obj))
	 * ; fprintf(stderr,"regs:%d\n",(tp->frames[tp->frame].cregs+1));
	 */
	memset(tp->frames[tp->frame].regs - TP_REGS_EXTRA, 0x0,
		(TP_REGS_EXTRA + tp->frames[tp->frame].cregs) * sizeof(tp_obj));
	tp->frame -= 1;
}

/* instruction opcodes */
enum {
	TP_IEOF = 0x00,
	TP_IADD = 0x01,
	TP_ISUB = 0x02,
	TP_IMUL = 0x03,
	TP_IDIV = 0x04,
	TP_IPOW = 0x05,
	TP_IBITAND = 0x06,
	TP_IBITOR = 0x07,
	TP_ICMP = 0x08,
	TP_IGET = 0x09,
	TP_ISET = 0x0a,
	TP_INUMBER = 0x0b,
	TP_ISTRING = 0x0c,
	TP_IGGET = 0x0d,
	TP_IGSET = 0x0e,
	TP_IMOVE = 0x0f,
	TP_IDEF = 0x10,
	TP_IPASS = 0x11,
	TP_IJUMP = 0x12,
	TP_ICALL = 0x13,
	TP_IRETURN = 0x14,
	TP_IIF = 0x15,
	TP_IDEBUG = 0x16,
	TP_IEQ = 0x17,
	TP_ILE = 0x18,
	TP_ILT = 0x19,
	TP_IDICT = 0x1a,
	TP_ILIST = 0x1b,
	TP_INONE = 0x1c,
	TP_ILEN = 0x1d,
	TP_ILINE = 0x1e,
	TP_IPARAMS = 0x1f,
	TP_IIGET = 0x20,
	TP_IFILE = 0x21,
	TP_INAME = 0x22,
	TP_INE = 0x23,
	TP_IHAS = 0x24,
	TP_IRAISE = 0x25,
	TP_ISETJMP = 0x26,
	TP_IMOD = 0x27,
	TP_ILSH = 0x28,
	TP_IRSH = 0x29,
	TP_IITER = 0x2a,
	TP_IDEL = 0x2b,
	TP_IREGS = 0x2c,
	TP_IBITXOR = 0x2d,
	TP_IIFN = 0x2e,
	TP_INOT = 0x2f,
	TP_IBITNOT = 0x30,
	TP_ITOTAL = 0x31
};

/*
 * char *tp_strings[TP_ITOTAL] = {
 * "EOF","ADD","SUB","MUL","DIV","POW","BITAND","BITOR","CMP","GET","SET","NUM
 * ",
 * "STR","GGET","GSET","MOVE","DEF","PASS","JUMP","CALL","RETURN","IF","DEBUG"
 * , "EQ","LE","LT","DICT","LIST","NONE","LEN","LINE","PARAMS","IGET","FILE",
 * "NAME","NE","HAS","RAISE","SETJMP","MOD","LSH","RSH","ITER","DEL","REGS",
 * "BITXOR", "IFN", "NOT", "BITNOT", };
 */

#define VA	((int)e.regs.a)
#define VB	((int)e.regs.b)
#define VC	((int)e.regs.c)
#define RA	regs[e.regs.a]
#define RB	regs[e.regs.b]
#define RC	regs[e.regs.c]
#define UVBC	(uint16_t)(((VB << 8) + VC))
#define SVBC	(int16_t)(((VB << 8) + VC))

#ifdef TP_USE_SWITCH
int 
tp_step(tp_vm * tp)
{
	tp_frame_      *f = &tp->frames[tp->frame];
	tp_obj         *regs = f->regs;
	tp_code        *in = f->in;
	tp_num		num;
	int		a;

	for (;;) {
#ifdef TP_SANDBOX
		tp_bounds(tp, in, 1);
#endif
		tp_code 	e = *in++;
		/*
	        fprintf(stderr,"%2d.%4d: %-6s %3d %3d %3d\n",
		 	tp->frame,
			in - (tp_code*)f->code.string.val,
			tp_strings[e.i],
			VA,
			VB,
			VC);
	        int i; for(i=0;i<16;i++) {
			fprintf(stderr,"%d: %s\n",i,TP_xSTR(regs[i]));
		}
	        */
		switch(e.i) {
		case TP_IEOF:
			tp_return(tp, tp_None);
			f->in = in - 1;
			return 0;
		case TP_IADD:
			RA = tp_add(tp, RB, RC);
			break;
		case TP_ISUB:
			RA = tp_sub(tp, RB, RC);
			break;
		case TP_IMUL:
			RA = tp_mul(tp, RB, RC);
			break;
		case TP_IDIV:
			RA = tp_div(tp, RB, RC);
			break;
		case TP_IPOW:
			RA = tp_pow(tp, RB, RC);
			break;
		case TP_IBITAND:
			RA = tp_bitwise_and(tp, RB, RC);
			break;
		case TP_IBITOR:
			RA = tp_bitwise_or(tp, RB, RC);
			break;
		case TP_IBITXOR:
			RA = tp_bitwise_xor(tp, RB, RC);
			break;
		case TP_IMOD:
			RA = tp_mod(tp, RB, RC);
			break;
		case TP_ILSH:
			RA = tp_lsh(tp, RB, RC);
			break;
		case TP_IRSH:
			RA = tp_rsh(tp, RB, RC);
			break;
		case TP_ICMP:
			RA = tp_number(tp_cmp(tp, RB, RC));
			break;
		case TP_INE:
			RA = tp_number(tp_cmp(tp, RB, RC) != 0);
			break;
		case TP_IEQ:
			RA = tp_number(tp_cmp(tp, RB, RC) == 0);
			break;
		case TP_ILE:
			RA = tp_number(tp_cmp(tp, RB, RC) <= 0);
			break;
		case TP_ILT:
			RA = tp_number(tp_cmp(tp, RB, RC) < 0);
			break;
		case TP_IBITNOT:
			RA = tp_bitwise_not(tp, RB);
			break;
		case TP_INOT:
			RA = tp_number(!tp_bool(tp, RB));
			break;
		case TP_IPASS:
			break;
		case TP_IIF:
			if (tp_bool(tp, RA)) {
				in += 1;
			}
			break;
		case TP_IIFN:
			if (!tp_bool(tp, RA)) {
				in += 1;
			}
			break;
		case TP_IGET:
			RA = tp_get(tp, RB, RC);
			tp_grey(tp,RA);
			break;
		case TP_IITER:
			if (RC.number.val < tp_len(tp, RB).number.val) {
				RA = tp_iter(tp, RB, RC);
				tp_grey(tp, RA);
				RC.number.val += 1;
#ifdef TP_SANDBOX
				tp_bounds(tp, in - 1, 1);
#endif
				in += 1;
			}
			break;
		case TP_IHAS:
			RA = tp_has(tp, RB, RC);
			break;
		case TP_IIGET:
			tp_iget(tp, &RA, RB, RC);
			break;
		case TP_ISET:
			tp_set(tp, RA, RB, RC);
			break;
		case TP_IDEL:
			tp_del(tp, RA, RB);
			break;
		case TP_IMOVE:
			RA = RB;
			break;
		case TP_INUMBER:
#ifdef TP_SANDBOX
			tp_bounds(tp, in - 1, sizeof(tp_num) / 4);
#endif
			memcpy(&num, in->string.val, sizeof(num));
			RA = tp_number(num);
			in += (sizeof(tp_num) / 4); // XXX
			break;
		case TP_ISTRING:
#ifdef TP_SANDBOX
			tp_bounds(tp, in - 1, (UVBC / 4) + 1);
#endif
			/*
			 * RA =
			 * tp_string_n((*(in+1)).string.val,UVBC);
			 */
			a = (*in).string.val - f->code.string.val;
			RA = tp_string_sub(tp, f->code, a, a + UVBC),
				in += (UVBC / 4) + 1;
			break;
		case TP_IDICT:
			RA = tp_dict_n(tp, VC / 2, &RB);
			break;
		case TP_ILIST:
			RA = tp_list_n(tp, VC, &RB);
			break;
		case TP_IPARAMS:
			RA = tp_params_n(tp, VC, &RB);
			break;
		case TP_ILEN:
			RA = tp_len(tp, RB);
			break;
		case TP_IJUMP:
			in += SVBC - 1;
			break;
		case TP_ISETJMP:
			f->jmp = SVBC ? in + SVBC - 1 : 0;
			break;
		case TP_ICALL:
#ifdef TP_SANDBOX
			tp_bounds(tp, in - 1, 1);
#endif
			f->in = in;
			RA = tp_call(tp, RB, RC);
			tp_grey(tp, RA);
			return 0;
		case TP_IGGET:
			if (!tp_iget(tp, &RA, f->globals, RB)) {
				RA = tp_get(tp, tp->builtins, RB);
				tp_grey(tp, RA);
			}
			break;
		case TP_IGSET:
			tp_set(tp, f->globals, RA, RB);
			break;
		case TP_IDEF:
			/*
			 * RA =
			 * tp_def(tp,(*(in+1)).string.val,f->globals)
			 * ;
			 */
#ifdef TP_SANDBOX
			tp_bounds(tp, in - 1, SVBC);
#endif
			a = (*in).string.val - f->code.string.val;
			RA = tp_def(tp,
			/*
			 * tp_string_n((*(in+1)).string.val,(SVBC-1)*4
			 * ),
			 */
				    tp_string_sub(tp, f->code, a, a + (SVBC - 1) * 4),
				    f->globals);
			in += SVBC - 1;
			break;
		case TP_IRETURN:
			tp_return(tp, RA);
			f->in = in - 1;
			return 0;
		case TP_IRAISE:
			_tp_raise(tp, RA);
			f->in = in - 1;
			return 0;
		case TP_IDEBUG:
			tp_params_v(tp, 3, tp_string("DEBUG:"), tp_number(VA), RA);
			tp_print(tp);
			break;
		case TP_INONE:
			RA = tp_None;
			break;
		case TP_ILINE:
#ifdef TP_SANDBOX
			tp_bounds(tp, in - 1, VA);
#endif
			a = (*in).string.val - f->code.string.val;
			/*
			 * f->line =
			 * tp_string_n((*(in+1)).string.val,VA*4-1);
			 */
			f->line = tp_string_sub(tp, f->code, a, a + VA * 4 - 1);
			/*
			 * fprintf(stderr,"%7d:
			 * %s\n",UVBC,f->line.string.val);
			 */
			in += VA;
			f->lineno = UVBC;
			break;
		case TP_IFILE:
			f->fname = RA;
			break;
		case TP_INAME:
			f->name = RA;
			break;
		case TP_IREGS:
			f->cregs = VA;
			break;
		default:
			tp_raise(0, tp_string("(tp_step) RuntimeError: invalid instruction"));
			break;
		}
#ifdef TP_SANDBOX
		tp_time_update(tp);
		tp_mem_update(tp);
		tp_bounds(tp, in - 1, 1);
#endif
	}
	f->in = in;
	return 0;
}
#else /* TP_USE_SWITCH */

#define DISPATCH()      						\
	tp_bounds(tp, in, 1);						\
	tp_time_update(tp);						\
	tp_mem_update(tp);						\
	e = *in++;							\
	goto *tab[(e.i >= TP_ITOTAL) ? TP_ITOTAL : e.i]

int 
tp_step(tp_vm * tp)
{
        static void     *tab[] = {
                &&lab_ieof, &&lab_iadd, &&lab_isub, &&lab_imul, &&lab_idiv,
                &&lab_ipow, &&lab_ibitand, &&lab_ibitor, &&lab_icmp, &&lab_iget,
                &&lab_iset, &&lab_inumber, &&lab_istring, &&lab_igget, &&lab_igset,
                &&lab_imove, &&lab_idef, &&lab_ipass, &&lab_ijump, &&lab_icall,
                &&lab_ireturn, &&lab_iif, &&lab_idebug, &&lab_ieq, &&lab_ile,
                &&lab_ilt, &&lab_idict, &&lab_ilist, &&lab_inone, &&lab_ilen, &&lab_iline,
                &&lab_iparams, &&lab_iiget, &&lab_ifile, &&lab_iname, &&lab_ine,
                &&lab_ihas, &&lab_iraise, &&lab_isetjmp, &&lab_imod, &&lab_ilsh,
                &&lab_irsh, &&lab_iiter, &&lab_idel, &&lab_iregs, &&lab_ibitxor,
		&&lab_iifn, &&lab_inot, &&lab_ibitnot, &&lab_iother
        };
	tp_frame_      *f;
	tp_code        *in;
	tp_code         e;
	tp_obj         *regs;
	tp_num		num;
	int		a;

	f = &tp->frames[tp->frame];
	regs = f->regs;
	in = f->in;
	DISPATCH();
lab_ieof:
	tp_return(tp, tp_None);
	f->in = in - 1;
	return 0;
lab_iadd:
	RA = tp_add(tp, RB, RC);
	DISPATCH();
lab_isub:
	RA = tp_sub(tp, RB, RC);
	DISPATCH();
lab_imul:
	RA = tp_mul(tp, RB, RC);
	DISPATCH();
lab_idiv:
	RA = tp_div(tp, RB, RC);
	DISPATCH();
lab_ipow:
	RA = tp_pow(tp, RB, RC);
	DISPATCH();
lab_ibitand:
	RA = tp_bitwise_and(tp, RB, RC);
	DISPATCH();
lab_ibitor:
	RA = tp_bitwise_or(tp, RB, RC);
	DISPATCH();
lab_ibitxor:
	RA = tp_bitwise_xor(tp, RB, RC);
	DISPATCH();
lab_imod:
	RA = tp_mod(tp, RB, RC);
	DISPATCH();
lab_ilsh:
	RA = tp_lsh(tp, RB, RC);
	DISPATCH();
lab_irsh:
	RA = tp_rsh(tp, RB, RC);
	DISPATCH();
lab_icmp:
	RA = tp_number(tp_cmp(tp, RB, RC));
	DISPATCH();
lab_ine:
	RA = tp_number(tp_cmp(tp, RB, RC) != 0);
	DISPATCH();
lab_ieq:
	RA = tp_number(tp_cmp(tp, RB, RC) == 0);
	DISPATCH();
lab_ile:
	RA = tp_number(tp_cmp(tp, RB, RC) <= 0);
	DISPATCH();
lab_ilt:
	RA = tp_number(tp_cmp(tp, RB, RC) < 0);
	DISPATCH();
lab_ibitnot:
	RA = tp_bitwise_not(tp, RB);
	DISPATCH();
lab_inot:
	RA = tp_number(!tp_bool(tp, RB));
	DISPATCH();
lab_ipass:
	DISPATCH();
lab_iif:
	if (tp_bool(tp, RA)) {
		in += 1;
	}
	DISPATCH();
lab_iifn:
	if (!tp_bool(tp, RA)) {
		in += 1;
	}
	DISPATCH();
lab_iget:
	RA = tp_get(tp, RB, RC);
	tp_grey(tp,RA);
	DISPATCH();
lab_iiter:
	if (RC.number.val < tp_len(tp, RB).number.val) {
		RA = tp_iter(tp, RB, RC);
		tp_grey(tp, RA);
		RC.number.val += 1;
		tp_bounds(tp, in - 1, 1);
		in += 1;
	}
	DISPATCH();
lab_ihas:
	RA = tp_has(tp, RB, RC);
	DISPATCH();
lab_iiget:
	tp_iget(tp, &RA, RB, RC);
	DISPATCH();
lab_iset:
	tp_set(tp, RA, RB, RC);
	DISPATCH();
lab_idel:
	tp_del(tp, RA, RB);
	DISPATCH();
lab_imove:
	RA = RB;
	DISPATCH();
lab_inumber:
	tp_bounds(tp, in - 1, sizeof(tp_num) / 4);
	memcpy(&num, in->string.val, sizeof(num));
	RA = tp_number(num);
	in += (sizeof(tp_num) / 4);
	DISPATCH();
lab_istring:
	tp_bounds(tp, in - 1, (UVBC / 4) + 1);
	a = (*in).string.val - f->code.string.val;
	RA = tp_string_sub(tp, f->code, a, a + UVBC),
		in += (UVBC / 4) + 1;
	DISPATCH();
lab_idict:
	RA = tp_dict_n(tp, VC / 2, &RB);
	DISPATCH();
lab_ilist:
	RA = tp_list_n(tp, VC, &RB);
	DISPATCH();
lab_iparams:
	RA = tp_params_n(tp, VC, &RB);
	DISPATCH();
lab_ilen:
	RA = tp_len(tp, RB);
	DISPATCH();
lab_ijump:
	in += SVBC - 1;
	DISPATCH();
lab_isetjmp:
	f->jmp = SVBC ? in + SVBC - 1 : 0;
	DISPATCH();
lab_icall:
	tp_bounds(tp, in - 1, 1);
	f->in = in;
	RA = tp_call(tp, RB, RC);
	tp_grey(tp, RA);
	return 0;
lab_igget:
	if (!tp_iget(tp, &RA, f->globals, RB)) {
		RA = tp_get(tp, tp->builtins, RB);
		tp_grey(tp, RA);
	}
	DISPATCH();
lab_igset:
	tp_set(tp, f->globals, RA, RB);
	DISPATCH();
lab_idef:
	tp_bounds(tp, in - 1, SVBC);
	a = (*in).string.val - f->code.string.val;
	RA = tp_def(tp,
		    tp_string_sub(tp, f->code, a, a + (SVBC - 1) * 4),
		    f->globals);
	in += SVBC - 1;
	DISPATCH();
lab_ireturn:
	tp_return(tp, RA);
	f->in = in - 1;
	return 0;
lab_iraise:
	_tp_raise(tp, RA);
	f->in = in - 1;
	return 0;
lab_idebug:
	tp_params_v(tp, 3, tp_string("DEBUG:"), tp_number(VA), RA);
	tp_print(tp);
	DISPATCH();
lab_inone:
	RA = tp_None;
	DISPATCH();
lab_iline:
	tp_bounds(tp, in - 1, VA);
	a = (*in).string.val - f->code.string.val;
	f->line = tp_string_sub(tp, f->code, a, a + VA * 4 - 1);
	in += VA;
	f->lineno = UVBC;
	DISPATCH();
lab_ifile:
	f->fname = RA;
	DISPATCH();
lab_iname:
	f->name = RA;
	DISPATCH();
lab_iregs:
	f->cregs = VA;
	DISPATCH();
lab_iother:
	tp_raise(0, tp_string("(tp_step) RuntimeError: invalid instruction"));
	DISPATCH();
}
#endif /* TP_USE_SWITCH */

static void 
internal_run(tp_vm * tp, int frame)
{
	tp->jmp += 1;
	if (setjmp(tp->buf)) {
		tp_handle(tp);
	}
	while (tp->frame >= frame && tp_step(tp) != -1) {
	}
	tp->jmp -= 1;
}

void 
tp_run(tp_vm * tp, int frame)
{
	jmp_buf 	tmp;

	memcpy(tmp, tp->buf, sizeof(jmp_buf));
	internal_run(tp, frame);
	memcpy(tp->buf, tmp, sizeof(jmp_buf));
}


tp_obj 
tp_ez_call(tp_vm * tp, const char *mod, const char *fnc, tp_obj params)
{
	tp_obj 		tmp;

	tmp = tp_get(tp, tp->modules, tp_string(mod));
	tmp = tp_get(tp, tmp, tp_string(fnc));
	return tp_call(tp, tmp, params);
}

static tp_obj 
internal_import(tp_vm * tp, tp_obj fname, tp_obj name, tp_obj code)
{
	tp_obj 		g;

	if (!((fname.type != TP_NONE && _tp_str_index(fname, tp_string(".tpc")) != -1) || code.type != TP_NONE)) {
		return tp_ez_call(tp, "py2bc", "import_fname", tp_params_v(tp, 2, fname, name));
	}
	if (code.type == TP_NONE) {
		tp_params_v(tp, 1, fname);
		code = tp_load(tp);
	}
	g = tp_dict(tp);
	tp_set(tp, g, tp_string("__name__"), name);
	tp_set(tp, g, tp_string("__code__"), code);
	tp_set(tp, g, tp_string("__dict__"), g);
	tp_frame(tp, g, code, 0);
	tp_set(tp, tp->modules, name, g);
	if (!tp->jmp) {
		tp_run(tp, tp->frame);
	}
	return g;
}


/*
 * Function: tp_import Imports a module.
 * 
 * Parameters: fname - The filename of a file containing the module's code. name
 * - The name of the module. codes - The module's code.  If this is given,
 * fname is ignored. len - The length of the bytecode.
 * 
 * Returns: The module object.
 */
tp_obj 
tp_import(tp_vm * tp, const char *fname, const char *name, void *codes, int len)
{
	tp_obj 		f = fname ? tp_string(fname) : tp_None;
	tp_obj 		bc = codes ? tp_string_n((const char *) codes, len) : tp_None;

	return internal_import(tp, f, tp_string(name), bc);
}



tp_obj 
tp_builtin_exec(tp_vm * tp)
{
	tp_obj 		code = TP_OBJ();
	tp_obj 		globals = TP_OBJ();
	tp_obj 		r = tp_None;

	tp_frame(tp, globals, code, &r);
	tp_run(tp, tp->frame);
	return r;
}


tp_obj 
tp_builtin_import(tp_vm * tp)
{
	tp_obj 		mod = TP_OBJ();
	tp_obj 		r;

	if (tp_has(tp, tp->modules, mod).number.val) {
		return tp_get(tp, tp->modules, mod);
	}
	r = internal_import(tp, tp_add(tp, mod, tp_string(".tpc")), mod, tp_None);
	return r;
}

void 
tp_builtins(tp_vm * tp)
{
	tp_obj 		o;
	struct {
		const char     *s;
		void           *f;
	} b[] = {
		{ "print", tp_print },
		{ "range", tp_range },
		{ "min", tp_min },
		{ "max", tp_max },
		{ "bind", tp_bind },
		{ "copy", tp_copy },
		{ "import", tp_builtin_import },
		{ "len", tp_builtin_len },
		{ "assert", tp_assert },
		{ "str", tp_builtin_str },
		{ "float", tp_float },
		{ "system", tp_system },
		{ "istype", tp_istype },
		{ "chr", tp_chr },
		{ "save", tp_save },
		{ "load", tp_load },
		{ "fpack", tp_fpack },
		{ "abs", tp_abs },
		{ "int", tp_int },
		{ "exec", tp_builtin_exec },
		{ "exists", tp_exists },
		{ "mtime", tp_mtime },
		{ "number", tp_float },
		{ "round", tp_round },
		{ "ord", tp_ord },
		{ "merge", tp_merge },
		{ "getraw", tp_getraw },
		{ "setmeta", tp_setmeta },
		{ "getmeta", tp_getmeta },
		{ "bool", tp_builtins_bool },
#ifdef TP_SANDBOX
		{ "sandbox", tp_builtin_sandbox },
#endif
		{ NULL, NULL },
	};
	int 		i;

	for (i = 0; b[i].s; i++) {
		tp_set(tp, tp->builtins, tp_string(b[i].s), tp_fnc(tp, (tp_obj(*) (tp_vm *)) b[i].f));
	}
	o = tp_object(tp);
	tp_set(tp, o, tp_string("__call__"), tp_fnc(tp, tp_object_call));
	tp_set(tp, o, tp_string("__new__"), tp_fnc(tp, tp_object_new));
	tp_set(tp, tp->builtins, tp_string("object"), o);
}


void 
tp_args(tp_vm * tp, int argc, char *argv[], int start)
{
	tp_obj 		self = tp_list(tp);
	int 		i;

	for (i = start; i < argc; i++) {
		_tp_list_append(tp, self.list.val, tp_string(argv[i]));
	}
	tp_set(tp, tp->builtins, tp_string("ARGV"), self);
}

tp_obj 
tp_main(tp_vm * tp, char *fname, void *code, int len)
{
	return tp_import(tp, fname, "__main__", code, len);
}

/*
 * Function: tp_compile Compile some tinypy code.
 * 
 */
tp_obj 
tp_compile(tp_vm * tp, tp_obj text, tp_obj fname)
{
	return tp_ez_call(tp, "BUILTINS", "compile", tp_params_v(tp, 2, text, fname));
}

/*
 * Function: tp_exec Execute VM code.
 */
tp_obj 
tp_exec(tp_vm *tp, tp_obj code, tp_obj globals)
{
	tp_obj 		r = tp_None;

	tp_frame(tp, globals, code, &r);
	tp_run(tp, tp->frame);
	return r;
}

tp_obj 
tp_eval(tp_vm *tp, const char *text, tp_obj globals)
{
	tp_obj 		code = tp_compile(tp, tp_string(text), tp_string("<eval>"));

	return tp_exec(tp, code, globals);
}

/*
 * Function: tp_init Initializes a new virtual machine.
 * 
 * The given parameters have the same format as the parameters to main, and
 * allow passing arguments to your tinypy scripts.
 * 
 * Returns: The newly created tinypy instance.
 */
tp_vm *
tp_init(int argc, char *argv[], int start)
{
	tp_vm          *tp = internal_init();

	tp_builtins(tp);
	tp_args(tp, argc, argv, start);
	return tp;
}

/**/
